# coding: utf-8
from __future__ import annotations

import asyncio
import platform
import sys
from dataclasses import dataclass, field
from typing import Any, Collection, Dict, List, Mapping, Type

import semver

from piaf.comm.mts import MessageTransportProtocol
from piaf.exceptions import InvalidStateException
from piaf.ptf import AgentPlatform, Extension, PlatformState


@dataclass(eq=True, frozen=True)
class ServiceDescription:
    """Description of a service to be launched by a :class:`PlatformLauncher` instance."""

    #: The :class:`AID` short name
    name: str

    #: The class
    clazz: Type

    #: A collection of arguments to be passed to the constructors
    args: Collection[Any] = field(default_factory=list)

    #: Same as  `args` but for keyword arguments
    kwargs: Mapping[str, Any] = field(default_factory=dict)


AgentDescription = ServiceDescription
AgentDescription.__doc__ = """Description of an agent to be launched by a :class:`PlatformLauncher` instance."""


class PlatformLauncher:
    """
    A simpler API to create, set up, launch and teardown a platform.

    Once the :class:`PlatformLauncher` instance is created, you can add agents, services and MTPs. Once ready, just
    call the :meth:`run` method. It will launch the platform and wait for you to hit Ctrl + C.
    """

    def __init__(self, ptf_name: str) -> None:
        """
        Create a new :class:`PlatformLauncher` instance.

        Initially, there is no agents, services or MTPs registered and the platform is created but not launched.

        :param ptf_name: name of the platform
        """
        self.ptf = AgentPlatform(ptf_name)
        self._agents: List[AgentDescription] = []
        self._services: List[ServiceDescription] = []
        self._mtps: List[MessageTransportProtocol] = []
        self._extensions: Dict[str, Extension] = {}

    def add_agent(self, agent_description: AgentDescription) -> None:
        """
        Add the given agent to the platform.

        :param agent_description: description of the agent to add
        :raise InvalidStateException: the platform is already launched
        """
        if self.ptf.state != PlatformState.INITIALIZED:
            raise InvalidStateException(self.ptf.state, "add_agent")
        self._agents.append(agent_description)

    def add_service(self, service_description: ServiceDescription) -> None:
        """
        Add the given service to the platform.

        :param service_description: description of the service to add
        :raise InvalidStateException: the platform is already launched
        """
        if self.ptf.state != PlatformState.INITIALIZED:
            raise InvalidStateException(self.ptf.state, "add_service")
        self._services.append(service_description)

    def add_mtp(self, mtp: MessageTransportProtocol) -> None:
        """
        Add the given MTP to the platform.

        :param mtp: the instantiated MTP
        :raise InvalidStateException: the platform is already launched
        """
        if self.ptf.state != PlatformState.INITIALIZED:
            raise InvalidStateException(self.ptf.state, "add_mtp")
        self._mtps.append(mtp)

    def add_extension(self, name: str, ext: Extension) -> None:
        """
        Add the given extension to the platform.

        :param name: name given for indexing
        :param extension: the extension to add
        :raise InvalidStateException: the platform is already launched
        """
        if self.ptf.state != PlatformState.INITIALIZED:
            raise InvalidStateException(self.ptf.state, "add_extension")
        self._extensions[name] = ext

    def run(self) -> None:
        """
        Configure and run the wrapped :class:`AgentPlatform`.

        This method takes care of getting the event loop and tearing down things. Control is given back with CTRL+C.
        """
        # Before Python 3.10.6, there is a bug in the ProactorEventLoop.
        # See https://github.com/nedbat/dinghy/issues/9 and https://github.com/python/cpython/issues/83413
        if semver.match(platform.python_version(), ">=3.8.0") and semver.match(
            platform.python_version(), "<3.10.6"
        ):

            def unraisablehook(unraisable, _old_hook=sys.unraisablehook):
                if (
                    unraisable.exc_value.args[0] == "Event loop is closed"
                    and unraisable.object.__name__ == "__del__"
                ):
                    return
                return _old_hook(unraisable)

            sys.unraisablehook = unraisablehook

        # Configure the loop and launch the platform
        loop = asyncio.get_event_loop()
        try:
            loop.create_task(self._main())
            loop.run_until_complete(self._wait_platform_death())
        except KeyboardInterrupt:
            pass
        finally:
            loop.run_until_complete(self._cleanup())
            tasks = asyncio.all_tasks(loop)
            if tasks:
                loop.run_until_complete(
                    asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
                )
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()

    async def _cleanup(self) -> None:
        # Stop the platform if ot done already
        if self.ptf.state != PlatformState.STOPPED:
            await self.ptf.stop()

        # Close the event manager since the platform doesn't manage its lifecycle
        await self.ptf.evt_manager.close()

    async def _main(self) -> None:
        """Coroutine that starts the platform and setup mtps, services and agents."""
        # Extensions are loaded first, before the platform is started since agents could rely on exposed primitives.
        for name, ext in self._extensions.items():
            self.ptf.load_extension(name, ext)

        # Before adding anything, we need to start the platform. This ensure that the
        # AMS agent is created
        await self.ptf.start()

        # Then register all provided MTP instances
        for mtp in self._mtps:
            await self.ptf.acc.register_mtp(mtp)

        # Then services
        for service in self._services:
            aid = await self.ptf.agent_manager.create(
                service.clazz, service.name, True, *service.args, **service.kwargs
            )
            await self.ptf.agent_manager.invoke(aid)

        # Then agents
        for agent in self._agents:
            aid = await self.ptf.agent_manager.create(
                agent.clazz, agent.name, True, *agent.args, **agent.kwargs
            )
            await self.ptf.agent_manager.invoke(aid)

    async def _wait_platform_death(self) -> None:
        """Wait until the platform is stopped."""
        async with self.ptf.state_sync:
            await self.ptf.state_sync.wait_for(
                lambda: self.ptf.state == PlatformState.STOPPED
            )



# coding: utf-8
"""This module contains application controllers."""
from __future__ import annotations

import asyncio
import json
import os
import signal
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Dict, List, Mapping
from uuid import uuid4

import async_timeout as at
from aioredis import ConnectionPool, Redis
from aioredis.client import PubSub
from aioredis.exceptions import RedisError as AIORedisError
from fastapi import status
from fastapi.responses import JSONResponse

from piaf.api.config import Settings
from piaf.api.exceptions import (
    DuplicatedPlatformNameError,
    InternalServerError,
    RedisConnectionNotInitialized,
    RedisError,
    RedisUnavailableError,
    UnknownPlatformError,
)
from piaf.api.impl.process import AgentPlatformProcess
from piaf.api.models import AgentPlatformModel


class RedisController:
    """
    Control a redis database.

    This controller maintains a connection pool bound to the Redis instance and can provide sessions (connections) to perform database operations.
    """

    def __init__(self, settings: Settings) -> None:
        """
        Create a new :class:`RedisController` instance.

        To bind the connections pool to the Redis database, you must call the :meth:`RedisController.connect` coroutine.

        :param settings: application settings, used to configure the connection pool
        """
        self._settings = settings
        self._pool: ConnectionPool | None = None

    async def connect(self) -> None:
        """
        Bind the connections pool to the Redis database and test it.

        :raise RedisUnavailableError: Can't reach the Redis database
        :raise RedisError: If any error occurs
        """
        options: Dict[str, Any] = {
            "decode_responses": True,
            "max_connections": self._settings.redis_max_connections,
        }
        if self._settings.redis_user and self._settings.redis_password:
            options.update(
                {
                    "username": self._settings.redis_user,
                    "password": self._settings.redis_password,
                }
            )
        if self._settings.redis_db:
            options["db"] = self._settings.redis_db

        self._pool = ConnectionPool.from_url(
            f"{self._settings.redis_scheme}://{self._settings.redis_host}", **options
        )
        async with self.session() as session:
            await session.ping()  # Ping to test the connection

    async def close(self) -> None:
        """Close the connections pool and disconnect all active connections."""
        if self._pool is not None:
            await self._pool.disconnect()

    @asynccontextmanager
    async def session(self) -> AsyncGenerator[Redis, None]:
        """
        Get a Redis session to the database.

        The session will be closed automatically after being used.

        :raise RedisConnectionNotInitialized: The connections pool is not bound, did you called :meth:`RedisController.connect` ?
        :raise RedisUnavailableError: The is an issue with the connection
        :raise RedisError: an error occurred while using the connection
        """
        if self._pool is None:
            raise RedisConnectionNotInitialized()

        try:
            session = Redis(connection_pool=self._pool)
            yield session
        except ConnectionError as e:
            raise RedisUnavailableError() from e
        except AIORedisError as e:
            raise RedisError() from e
        finally:
            if session is not None:
                await session.close()

    def __eq__(self, other: Any) -> bool:
        """
        Equality check.

        :param other: the other object
        :return: `True` if both are equal (same pool and settings), `False` otherwise.
        """
        if not isinstance(other, type(self)):
            return False
        return self._pool == other._pool and self._settings == other._settings

    def __hash__(self) -> int:
        """Get this instance's hash code."""
        return hash((self._pool, self._settings))


class PlatformController:
    """
    Controls platforms.

    This controller will allow you to manipulate platforms : creation, deletion, listing and sending tasks.
    """

    PLATFORMS_REDIS_SET = "platforms"
    PLATFORM_PREFIX_HASHSET = "platform:"

    def __init__(self, settings: Settings, session: Redis) -> None:
        """
        Create a new :class:`PlatformController`.

        :param settings: application settings to configure new platforms.
        :param session: an active Redis session
        """
        self._settings = settings
        self._session = session

    async def create_platform(self, ptf: AgentPlatformModel) -> None:
        """
        Try to spawn a new platform.

        :param ptf: the platform description supplied by the user
        :raise DuplicatedPlatformName: the supplied platform's name is already taken
        """
        if await self._session.sismember(self.PLATFORMS_REDIS_SET, ptf.name):
            raise DuplicatedPlatformNameError(ptf.name)

        # Subscribe to events coming from the future platform
        try:
            pubsub: PubSub = self._session.pubsub(ignore_subscribe_messages=True)
            await pubsub.subscribe(f"channels:{ptf.name}:events:.platform.agents.api")

            # Launch the platform
            info = self._spawn_process(ptf)

            # Wait until the platform is ready (ie the APIAgent is up)
            async for msg in pubsub.listen():
                record: Dict[str, Any] = json.loads(msg["data"])
                if record["event"]["type"] == "agent_creation":
                    break
        finally:
            await pubsub.close()

        # Store information about the platform in the redis database
        await self._session.sadd(self.PLATFORMS_REDIS_SET, ptf.name)
        await self._session.hset(self.PLATFORM_PREFIX_HASHSET + ptf.name, mapping=info)

    def _spawn_process(self, ptf: AgentPlatformModel) -> Dict[str, Any]:
        """
        Spawn a new agent platform as a process.

        :param ptf: description of the platform to create
        :return: a mapping describing the platform : the name, the type (process) and the process's pid
        """
        process = AgentPlatformProcess(ptf.name, self._settings)
        process.start()

        return {"type": "process", "name": ptf.name, "pid": process.pid}

    def _spawn_docker(self, ptf: AgentPlatformModel) -> Dict[str, Any]:
        """
        Spawn a new agent platform as a docker container.

        :param ptf: description of the platform to create
        :return: a mapping describing the platform : the name, the type (docker) and the container's hash
        """
        raise NotImplementedError()

    async def stop(self, name: str) -> None:
        """
        Stop and delete the platform identified by the provided description.

        :param name: the platform's name
        :raise UnknownPlatformError: there is no platform matching the provided description
        """
        if not (await self._session.sismember(self.PLATFORMS_REDIS_SET, name)):
            raise UnknownPlatformError(name)

        try:
            # Subscribe to platform events
            pubsub: PubSub = self._session.pubsub(ignore_subscribe_messages=True)
            await pubsub.subscribe(f"channels:{name}:events:.platform")

            # Publish the task to stop the platform
            await self._session.publish(
                f"channels:{name}:from-api",
                json.dumps({"task_type": "StopPlatformTask", "id": str(uuid4())}),
            )

            # Wait until a 'platform-death' event is received
            async for msg in pubsub.listen():
                record: Dict[str, Any] = json.loads(msg["data"])
                if (
                    record["event"]["type"] == "state_change"
                    and record["event"]["data"]["to"] == "STOPPED"
                ):
                    break
        except Exception as e:
            raise InternalServerError() from e
        finally:
            await pubsub.close()

        # Delete record from database
        record = await self._session.hgetall(self.PLATFORM_PREFIX_HASHSET + name)
        await self._session.delete(self.PLATFORM_PREFIX_HASHSET + name)
        await self._session.srem(self.PLATFORMS_REDIS_SET, name)

    async def kill(self, name: str) -> None:
        """
        Kill the provided platform without letting it finishing properly.

        :param name: name of the platform
        :raise UnknownPlatformError: there is no platform matching the provided description
        """
        if not (await self._session.sismember(self.PLATFORMS_REDIS_SET, name)):
            raise UnknownPlatformError(name)

        # Get the platform's description
        record: Dict[str, Any] = await self._session.hgetall(
            self.PLATFORM_PREFIX_HASHSET + name
        )
        if record["type"] == "process":
            await self._kill_process(record)
        else:
            await self._kill_docker(record)

        # Delete record from database
        await self._session.delete(self.PLATFORM_PREFIX_HASHSET + name)
        await self._session.srem(self.PLATFORMS_REDIS_SET, name)

    async def _kill_process(self, description: Dict[str, Any]) -> None:
        """
        Kill a platform running in a dedicated process.

        :param description: description of the platform to kill. It should contain a 'type' entry set to 'process' and a 'pid' entry set to the process's pid.
        """
        os.kill(int(description["pid"]), signal.SIGTERM)

    async def _kill_docker(self, description: Dict[str, Any]) -> None:
        """
        Kill a platform running in a docker container.

        :param description: description of the platform to kill. It should contain a 'type' entry set
        to 'docker' and a 'hash' entry set to the container's hash.
        """
        raise NotImplementedError()

    async def get_all_platforms(self) -> List[Mapping[str, str]]:
        """
        Interrogate the Redis database to get all the running platforms.

        :return: a list of platform names
        """
        records = await self._session.smembers(self.PLATFORMS_REDIS_SET)
        return [{"name": ptf} for ptf in records]

    async def process_task(
        self,
        ptf_name: str,
        task: Dict[str, Any],
        timeout: float = 5,
    ) -> JSONResponse:
        """
        Send the given task to the APIAgent in the specified platform.

        If the platform doesn't exists, then returns an HTTP 404 response. Otherwise the task is submitted to the platform and the result is returned.

        :param ptf_name: name of the platform
        :param task: a JSON-compatible dict describing the task to submit
        :param platform_ctrl: the platform controller
        :param timeout: time (in seconds) given to the task for its execution
        :return: a :class:`JSONRequest` with the result or the error.
        """
        # Ensure platform exists
        platforms = await self.get_all_platforms()
        if ptf_name not in (ptf["name"] for ptf in platforms):
            return JSONResponse(
                content={"details": f"Unknown platform '{ptf_name}'"},
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        # Subscribe
        pubsub = self._session.pubsub(ignore_subscribe_messages=True)
        await pubsub.subscribe(f"channels:{ptf_name}:to-api")

        # Wait for reply
        try:
            async with at.timeout(timeout):

                # Send task here (retry until delivered to the API agent or timeout)
                while (
                    await self._session.publish(
                        f"channels:{ptf_name}:from-api", json.dumps(task)
                    )
                    == 0
                ):
                    await asyncio.sleep(0.1)

                # Await response
                async for msg in pubsub.listen():
                    reply = json.loads(msg["data"])
                    if reply["id"] == task["id"]:
                        break
        except Exception as e:
            raise InternalServerError() from e
        finally:
            await pubsub.close()

        # Return result as a valid response
        if reply["error"] is not None:
            raise InternalServerError() from reply["error"]
        return JSONResponse(reply["data"])



# coding: utf-8
"""This module enables communication between platforms using the AMQP 0.9.1 protocol."""
from __future__ import annotations

import asyncio
import logging
import pickle

try:
    import aiormq
    import aiormq.abc
    from aiormq import PublishError
    from yarl import URL
except ImportError as e:
    print("Install extra amqp-mtp to use amqp message transfer protocol.")
    raise e

from piaf.comm import Message
from piaf.comm.mts import MessageContext, MessageTransportProtocol


class AMQPMessageTransportProtocol(MessageTransportProtocol):
    """
    A MTP implementation enabling communications using the AMQP 0.9.1 protocol.

    This MTP is built on asyncio and runs on the same thread as the platform. It first
    establishes a connection with an AMQP server and then:

        - use (and create if necessary) an exchange called "acc" in direct mode
        - bind an unnamed queue on it, using the platform name as routing key

    Messages are published on the "acc" exchange, using the destination platform name as routing key.
    """

    def __init__(self, amqp_server_url: str) -> None:
        """
        Initialize a new instance of :class:`AMQPMessageTransferProtocol`.

        :param amqp_server_url: URL to the AMQP server this connector will connect to.
        """
        super().__init__()
        self._server_url = amqp_server_url
        self._listening_condition = asyncio.Condition()
        self._listening_flag = False
        self.logger = logging.getLogger(type(self).__name__)

    @property
    def scheme(self) -> str:
        """
        Get the scheme of the protocol.

        :return: the scheme
        """
        return "amqp"

    @property
    def address(self) -> str:
        """
        Get the ACC address using the current protocol.

        :return: URL of the platform's ACC using the current protocol
        """
        return f"{self.scheme}://{self.acc.ptf_name}/acc"

    @property
    def type(self) -> str:
        """
        Get the type of this protocol (what to put in `protocol` fields).

        :return: the protocol type
        """
        return "piaf.mts.mtp.amqp.aio"

    async def send(self, message: Message, address: str) -> bool:
        """
        Send the provided message to the provided address (expected to be an amqp one).

        :param message: the message to send
        :param address: the amqp address of the other platform's ACC
        :return: `True` if the message is sent, `False` if the provided address is not an AMQP
            one or the MTP is unable to deliver the message to the remote platform.
        """
        parsed_url = URL(address)
        if parsed_url.scheme != self.scheme:
            return False

        rt_key = parsed_url.host
        if rt_key is None:
            return False

        # Make sure we are ready to receive responses
        if not self._listening_flag:
            async with self._listening_condition:
                await self._listening_condition.wait()

        try:
            await self.channel.basic_publish(
                body=pickle.dumps(message, protocol=4),
                exchange="acc",
                routing_key=rt_key,
                mandatory=True,
            )
        except PublishError as e:
            self.logger.warn("[AMQP] Unable to route message.", exc_info=e)
            return False

        self.logger.debug("[AMQP] Sent %s", message)
        return True

    async def start(self) -> None:
        """
        Start this MTP.

        It wil establish the connection to the AMQP server and create a long running task to
        listen to incoming messages.
        """
        await self._connect()
        self.listener = asyncio.create_task(self._listen())

    async def stop(self) -> None:
        """
        Stop this MTP.

        Cancel the listening task and close the connection with the AMQP server.
        """
        self.listener.cancel()
        await self.channel.close()
        await self.connection.close()

    async def _connect(self) -> None:
        """Establish the connection with the AMQP server and create a channel."""
        # Create channel
        self.connection = await aiormq.connect(URL(self._server_url))
        self.channel = await self.connection.channel()

    async def _listen(self) -> None:
        """A long running task that listens on incoming messages and transfer them to the ACC."""
        # Declare exchange & random queue
        await self.channel.exchange_declare(exchange="acc", exchange_type="direct")
        declare_ok = await self.channel.queue_declare(
            durable=False, exclusive=True, auto_delete=True
        )

        # Bind queue to this platform and start listening
        await self.channel.queue_bind(
            declare_ok.queue, "acc", routing_key=self.acc.ptf_name
        )

        # Set listening flag to true
        async with self._listening_condition:
            self._listening_flag = True
            self._listening_condition.notify_all()
        await self.channel.basic_consume(declare_ok.queue, self._on_message)

    async def _on_message(self, msg: aiormq.abc.DeliveredMessage) -> None:
        """
        Callback called when an incoming message is received.

        :param msg: the received message
        """
        self.logger.debug("[AMQP] Received %s", pickle.loads(msg.body))
        self.acc.forward(pickle.loads(msg.body), MessageContext("", self))



# coding: utf-8
# /usr/bin/env python3
"""
The :mod:`piaf.service` module defines a set of services the platform may offer to running agents.

Except the :class:`AMSService` service, if you want to activate a service you have to start it like a regular agent.
The only key difference is that services have a privileged access to the platform. To enable this special access, you
must set the `is_service` parameter in :meth:`piaf.ptf.AgentManager.create` method to `True`.
"""
from __future__ import annotations

import importlib
from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Sequence, Set, Tuple

from piaf.agent import Agent, AgentState
from piaf.behavior import Behavior
from piaf.comm import AID, MT_NOT, MT_OR, MT_PERFORMATIVE, ACLMessage, Performative
from piaf.exceptions import DuplicatedNameException
from piaf.util import *

if TYPE_CHECKING:
    import piaf.comm
    import piaf.ptf


class AMSService(Agent):
    """
    Special agent providing Agent Management Services.

    The AMS is a required component of the agent platform. It allows others agents to query the platform capabilities
    and other agents AIDs.

    The AMS uses the underlying platform to discover registered agents.

    .. warning:: For now, the AMS agent only supports the search, the modify and the create-agent requests.

    Specification: http://fipa.org/specs/fipa00023/SC00023K.html
    """

    SEARCH_FUNC = "search"
    CREATE_AGENT_FUNC = "create_agent"
    MODIFY_FUNC = "modify"

    def __init__(self, aid: piaf.comm.AID, platform: piaf.ptf.AgentPlatformFacade):
        """
        Create a new AMS Agent with the provided information.

        :param aid: this agent's identifier
        :param platform: the platform where this AMS agent is deployed
        """
        super().__init__(aid, platform)

        # Add behaviors
        self.add_behavior(_AMSRequestBehavior(self))
        self.add_behavior(_HandleInvalidMessageBehavior(self))


class _BaseRequestBehavior(FIPARequestProtocolBehavior, metaclass=ABCMeta):
    """Base implementation for request protocol handling by both the  AMS and the DF."""

    #: Acceptable messages for this Request protocol
    VALID_MSG_TEMPLATE = MT_OR(
        MT_PERFORMATIVE(Performative.REQUEST), MT_PERFORMATIVE(Performative.CANCEL)
    )

    def check_message(self, msg: ACLMessage) -> bool:
        """
        Check the content of the provided message.

        * Not a sequence      -> NOT_UNDERSTOOD, unsupported-value
        * Length == 0         -> REFUSE, missing-parameter
        * First param not str -> REFUSE, unrecognized-parameter-value

        :param msg: message to check
        :return: True if message is ok, False otherwise
        """
        content = msg.content
        result = False

        if not isinstance(content, (List, tuple)):
            self.agent.send(
                not_understood_message_from_request(msg, "Unsupported value: content")
            )

        elif len(content) == 0:
            self.agent.send(
                refuse_message_from_request(msg, "Missing parameter: function_name")
            )

        elif not isinstance(content[0], str):
            self.agent.send(
                refuse_message_from_request(
                    msg, f"Unrecognized parameter value: function_name, {content[0]}"
                )
            )

        else:
            result = True

        return result

    @property
    def msg_template(self) -> piaf.comm.MessageTemplate:
        return _BaseRequestBehavior.VALID_MSG_TEMPLATE


class _AMSRequestBehavior(_BaseRequestBehavior):
    """Custom implementation of the class:`FIPARequestProtocolBehavior` for the AMS agent."""

    async def on_valid_request(self, msg: ACLMessage) -> None:
        """
        Depending on the message content, the AMS agent is executing the request.

        For now, only the SEARCH function is supported.

        :param msg: the request message to handle
        """
        # If function is the search function
        if msg.content[0] == AMSService.SEARCH_FUNC:
            await self._handle_search_request(msg)

        # If function is the create-agent function
        elif msg.content[0] == AMSService.CREATE_AGENT_FUNC:
            await self._handle_create_agent_request(msg)

        # If function is the modify function
        elif msg.content[0] == AMSService.MODIFY_FUNC:
            await self._handle_modify_request(msg)

        # Unsupported function
        else:
            self.agent.send(
                refuse_message_from_request(
                    msg, f"Unsupported function: {msg.content[0]}"
                )
            )

    async def _handle_search_request(self, msg: ACLMessage) -> None:
        """Execute the `search` function if possible and send the reply."""
        # At least the mandatory parameter ams-agent-description
        if len(msg.content) < 2:
            self.agent.send(
                refuse_message_from_request(
                    msg, "Missing argument: ams-agent-description"
                )
            )
            return

        # At most 2 parameters : ams-agent-description and the optional parameter search-constraints
        if len(msg.content) > 3:
            self.agent.send(
                refuse_message_from_request(msg, "Unexpected argument count")
            )
            return

        # ams-agent-description must be a AMSAgentDescription
        if not isinstance(msg.content[1], AMSAgentDescription):
            self.agent.send(
                refuse_message_from_request(
                    msg,
                    f"Unrecognised parameter value: ams-agent-description, {msg.content[1]}",
                )
            )
            return

        # If present, search-constraints parameter must be a SearchConstraints
        if len(msg.content) == 3 and not isinstance(msg.content[2], SearchConstraints):
            self.agent.send(
                refuse_message_from_request(
                    msg,
                    f"Unrecognised parameter value: search-constraints, {msg.content[2]}",
                )
            )
            return

        ams_agent_description = msg.content[1]
        am = self.agent._platform.agent_manager
        agents = am.get_agents(ams_agent_description.state)
        agt_descriptions = []

        # Perform the search
        self.agent.send(agree_message_from_request(msg))

        if ams_agent_description.name is not None:
            if ams_agent_description.name in agents:
                agt_descriptions.append(
                    AMSAgentDescription(
                        ams_agent_description.name,
                        None,
                        am.get_state(ams_agent_description.name),
                    )
                )

        else:
            agt_descriptions = [
                AMSAgentDescription(agent, None, am.get_state(agent))
                for agent in agents
            ]

        # Constrain if applicable
        if len(msg.content) == 3:
            max_results = msg.content[2].max_results
            if max_results is not None:
                agt_descriptions = agt_descriptions[:max_results]

        # Send the result
        self.agent.send(
            inform_message_from_request(msg, [msg.content, agt_descriptions])
        )

    async def _handle_create_agent_request(self, msg: ACLMessage) -> None:
        # At least the mandatory parameter agent-creation-description
        if len(msg.content) < 2:
            self.agent.send(
                refuse_message_from_request(
                    msg, "Missing argument: agent-creation-description"
                )
            )
            return

        # At most 1 parameter : agent-creation-description
        if len(msg.content) > 2:
            self.agent.send(
                refuse_message_from_request(msg, "Unexpected argument count")
            )
            return

        # agent-creation-description must be a AgentCreationDescription
        if not isinstance(msg.content[1], AgentCreationDescription):
            self.agent.send(
                refuse_message_from_request(
                    msg,
                    f"Unrecognised parameter value: agent-creation-description, {msg.content[1]}",
                )
            )
            return

        # Agree request
        self.agent.send(agree_message_from_request(msg))

        # Try to instantiate the class
        description: AgentCreationDescription = msg.content[1]

        class_path = description.class_name.split(".")
        if len(class_path) <= 1:
            self.agent.send(failure_message_from_request(msg, "No module provided."))
            return

        module_name = ".".join(class_path[:-1])
        class_name = class_path[-1]

        try:
            module = importlib.import_module(module_name)
            agent_class = getattr(module, class_name)
        except ModuleNotFoundError:
            self.agent.send(
                failure_message_from_request(msg, f"No module named '{module_name}'")
            )
            return
        except AttributeError:
            self.agent.send(
                failure_message_from_request(
                    msg,
                    f"No agent class named '{class_name}' in module '{module_name}'",
                )
            )
            return

        try:
            if description.args is not None:
                aid = await self.agent._platform.agent_manager.create(
                    agent_class,
                    description.agent_name,
                    description.is_service,
                    description.args,
                )
            else:
                aid = await self.agent._platform.agent_manager.create(
                    agent_class,
                    description.agent_name,
                    description.is_service,
                )
            await self.agent._platform.agent_manager.invoke(aid)
            self.agent.send(inform_message_from_request(msg, [msg.content, aid]))
        except DuplicatedNameException as e:
            self.agent.send(
                failure_message_from_request(
                    msg,
                    str(e),
                )
            )
        except Exception as e:
            self.agent.send(
                failure_message_from_request(msg, f"Internal error: {str(e)}")
            )

    async def _handle_modify_request(self, msg: ACLMessage) -> None:
        """Handle an incoming modification request."""
        content: List[str | Any] = msg.content
        if len(content) < 2:
            self.agent.send(
                refuse_message_from_request(
                    msg, "Missing parameter: ams-agent-description"
                )
            )
            return

        if len(content) > 2:
            self.agent.send(
                refuse_message_from_request(msg, "Unexpected argument count")
            )
            return

        ams_agent_description = content[1]
        if not isinstance(ams_agent_description, AMSAgentDescription):
            self.agent.send(
                refuse_message_from_request(
                    msg,
                    f"Unrecognised parameter value: agent-creation-description, {ams_agent_description}",
                )
            )
            return

        self.agent.send(agree_message_from_request(msg))

        # This feature is deactivated because otherwise the modify function has very little interest.
        # Once the ownership is implemented, it could become interesting to limit agent state modification to both the agent itself and its owner
        # if (
        #     ams_agent_description.name is not None
        #     and ams_agent_description.name != msg.sender
        # ):
        #     self.agent.send(failure_message_from_request(msg, "Unauthorized"))
        #     return

        if ams_agent_description.ownership is not None:
            self.agent.send(failure_message_from_request(msg, "Unsupported: ownership"))
            return

        state = ams_agent_description.state

        if state is None:
            self.agent.send(failure_message_from_request(msg, "Missing new state"))
            return

        if state not in [
            AgentState.ACTIVE,
            AgentState.SUSPENDED,
            AgentState.UNKNOWN,
        ]:
            self.agent.send(
                failure_message_from_request(msg, f"Unsupported state: {state}")
            )
            return

        am: piaf.ptf.AgentManager = self.agent._platform.agent_manager

        # Compute AID (if omitted, use the sender)
        aid: AID = ams_agent_description.name  # type: ignore
        if aid is None:
            aid = msg.sender

        try:
            if state == AgentState.ACTIVE:
                await am.resume(aid)
            elif state == AgentState.SUSPENDED:
                await am.suspend(aid)
            else:
                await am.quit(aid)
            self.agent.send(inform_message_from_request(msg, msg.content))
        except Exception as e:
            self.agent.send(failure_message_from_request(msg, str(e)))


class _HandleInvalidMessageBehavior(Behavior):
    """
    Handle all messages not matching the template defined in :class:`_BaseRequestBehavior`.

    If such message is found then this behavior sends a `NOT_UNDERSTOOD` message.
    """

    def done(self) -> bool:
        """Infinite behavior."""
        return False

    async def action(self) -> None:
        """Wait for messages and send `NOT_UNDERSTOOD` message."""
        msg = await self.agent.receive(MT_NOT(_BaseRequestBehavior.VALID_MSG_TEMPLATE))
        self.agent.send(
            not_understood_message_from_request(
                msg.acl_message, f"Unsupported Act: {msg.acl_message.performative}"
            )
        )


@dataclass(eq=True, frozen=True)
class AMSAgentDescription:
    """
    :class:`AMSAgentDescription` objects are returned when querying the AMS agent about agents in the platform.

    Fields are:

    - **name**: the :class:`AID` of the agent
    - **ownership**: who owns it
    - **state**: the agent current state (at the time the AMS queried it)

    It is part of the fipa-agent-management ontology. See http://fipa.org/specs/fipa00023/SC00023K.html.
    """

    name: None | AID = None
    ownership: None | str = None
    state: None | AgentState = None


@dataclass(eq=True, frozen=True)
class AgentPlatformService:
    """
    Description of a platform service.

    Fields are:

    - **name**: name of the service
    - **type**: type of the service. For example: "fipa.mtp.http"
    - **addresses**: a sequence of urls where to access the service

    It is part of the fipa-agent-management ontology. See http://fipa.org/specs/fipa00023/SC00023K.html.
    """

    name: str
    type: str
    addresses: Sequence[str]


@dataclass(eq=True, frozen=True)
class AgentPlatformDescription:
    """
    Description of the AgentPlatform.

    Fields are:

    - **name**: the name of the agent platform
    - **ap_services**: a set of services this platform provides

    It is part of the fipa-agent-management ontology. See http://fipa.org/specs/fipa00023/SC00023K.html.
    """

    name: str
    ap_services: Set[AgentPlatformService]


@dataclass(eq=True, frozen=True)
class Property:
    """
    :class:`Property` objects are usefull for specifying parameter/value pairs.

    Part of the fipa-agent-management ontology. See http://fipa.org/specs/fipa00023/SC00023K.html.
    """

    name: str
    value: Any


@dataclass(eq=True, frozen=True)
class AgentCreationDescription:
    """
    Each object of type :class:`AgentCreationDescription` describes how to create and invoke an agent using the AMS service.

    .. note:: Not part of the official fipa-agent-management ontology but rather an extension to support agent creation like
        described in the specification at http://fipa.org/specs/fipa00023/SC00023K.html.
    """

    class_name: str
    agent_name: str
    args: List[Any] | None = None
    is_service: bool = True


class DFService(Agent):
    """
    The DFService acts like yellow pages

    This service stores agent capabilities on demand (meaning that each agent must register itself to this service). It
    supports all operations defined in the fipa specification :

    - register: register a new agent description
    - deregister: remove an agent description
    - search: perform a search (with a search pattern and constraints) against all records
    - modify: modify an existing record

    .. note:: The DFService currently doesn't support the federated DF mechanism desfined in fipa specification.
    """

    REGISTER_FUNC = "register"
    DEREGISTER_FUNC = "deregister"
    SEARCH_FUNC = "search"
    MODIFY_FUNC = "modify"

    def __init__(self, aid: piaf.comm.AID, platform: piaf.ptf.AgentPlatform):
        super().__init__(aid, platform)

        # Store agents and services descriptions
        self.storage: Dict[AID, DFAgentDescription] = {}

        self.add_behavior(_HandleInvalidMessageBehavior(self))
        self.add_behavior(_DFRequestBehavior(self))


class _DFRequestBehavior(_BaseRequestBehavior):
    async def on_valid_request(self, msg: ACLMessage) -> None:
        """
        Depending on the message content, the DF agent is executing the request.

        For now, only the REGISTER function is supported.

        :param msg: the request message to handle
        """
        # If function is the register function
        if msg.content[0] == DFService.REGISTER_FUNC:
            await self._handle_registration_request(msg, True)

        # If function is the deregister function
        elif msg.content[0] == DFService.DEREGISTER_FUNC:
            await self._handle_registration_request(msg, False)

        # If function is the search function
        elif msg.content[0] == DFService.SEARCH_FUNC:
            await self._handle_search_request(msg)

        # If function is the modify function
        elif msg.content[0] == DFService.MODIFY_FUNC:
            await self._handle_modify_request(msg)

        # Unsupported function
        else:
            self.agent.send(
                refuse_message_from_request(
                    msg, f"Unsupported function: {msg.content[0]}"
                )
            )

    async def _handle_registration_request(
        self, msg: ACLMessage, register: bool
    ) -> None:
        """
        Handle a register / deregister request.

        :param msg: the request
        :param register: whether this is a register request (`True`)  or a deregister one (`False`)
        """
        # Expected arity is 1
        try:
            df_agt_description = msg.content[1]
        except IndexError:
            self.agent.send(
                refuse_message_from_request(
                    msg, "Missing argument: df agent description"
                )
            )
            return

        # Parameter type should be DFAgentDescription
        if not isinstance(df_agt_description, DFAgentDescription):
            self.agent.send(
                refuse_message_from_request(
                    msg,
                    f"Unrecognised parameter value: df agent description, {df_agt_description}",
                )
            )
            return

        # Message is well formed, we can handle it
        self.agent.send(agree_message_from_request(msg))

        if register:
            # Agent must not be registered
            if df_agt_description.name in self.agent.storage:
                self.agent.send(failure_message_from_request(msg, "Already registered"))
                return

            # Everything is ok, we can process the registration
            self.agent.storage[df_agt_description.name] = df_agt_description
            self.agent.send(inform_message_from_request(msg, msg.content))
        else:
            # Agent must be registered
            if df_agt_description.name not in self.agent.storage:
                self.agent.send(failure_message_from_request(msg, "Not registered"))
                return

            # Everything is ok, we can process the deregister request
            del self.agent.storage[df_agt_description.name]
            self.agent.send(inform_message_from_request(msg, msg.content))

    async def _handle_search_request(self, msg: ACLMessage) -> None:
        """
        Handle a search request.

        Message content must contains at least one parameter called df-agent-description and eventually another one
        called search-constraints. If the conditions are not met, then a REFUSE message is sent with the appropriate
        error class.

        Otherwise, two messages are sent: an AGREE, followed by an inform containing the result.

        This implementation doesn't support the federated DF feature for now. As a result, max-depth and search-id
        constraints are ignored.

        :param msg: message containing the search request
        """
        # At least the mandatory parameter df-agent-description
        if len(msg.content) < 2:
            self.agent.send(
                refuse_message_from_request(
                    msg, "Missing argument: df-agent-description"
                )
            )
            return

        # At most 2 parameters : df-agent-description and the optional parameter search-constraints
        if len(msg.content) > 3:
            self.agent.send(
                refuse_message_from_request(msg, "Unexpected argument count")
            )
            return

        # df-agent-description must be a DFAgentDescription
        if not isinstance(msg.content[1], DFAgentDescription):
            self.agent.send(
                refuse_message_from_request(
                    msg,
                    f"Unrecognised parameter value: df-agent-description, {msg.content[1]}",
                )
            )
            return

        # If present, search-constraints parameter must be a SearchConstraints
        if len(msg.content) == 3 and not isinstance(msg.content[2], SearchConstraints):
            self.agent.send(
                refuse_message_from_request(
                    msg,
                    f"Unrecognised parameter value: search-constraints, {msg.content[2]}",
                )
            )
            return

        # Everything is fine, we can handle the request
        self.agent.send(agree_message_from_request(msg))

        result = self._filter_descriptions(msg.content[1])

        # Constraint result
        if len(msg.content) == 3:
            constraints: "SearchConstraints" = msg.content[2]
            if constraints.max_results is not None:
                result = result[: constraints.max_results]

        self.agent.send(inform_message_from_request(msg, [msg.content, result]))

    async def _handle_modify_request(self, msg: ACLMessage) -> None:
        """
        Handle a modify request.

        Expect one argument named df-agent-description of type :class:`DFAgentDescription` which is the new description
        for the sender agent. If an agent tries to modify the description of an other agent, a failure message will be
        sent.

        :param msg: modify request
        """
        # At least the mandatory parameter df-agent-description
        if len(msg.content) < 2:
            self.agent.send(
                refuse_message_from_request(
                    msg, "Missing argument: df-agent-description"
                )
            )
            return

        # At most 2 parameters : df-agent-description and the optional parameter search-constraints
        if len(msg.content) > 2:
            self.agent.send(
                refuse_message_from_request(msg, "Unexpected argument count")
            )
            return

        # df-agent-description must be a DFAgentDescription
        if not isinstance(msg.content[1], DFAgentDescription):
            self.agent.send(
                refuse_message_from_request(
                    msg,
                    f"Unrecognised parameter value: df-agent-description, {msg.content[1]}",
                )
            )
            return

        # Everything is fine, we can handle the request
        self.agent.send(agree_message_from_request(msg))

        # If sender is modifying description of another agent, failure
        if msg.content[1].name != msg.sender:
            self.agent.send(failure_message_from_request(msg, "Unauthorized"))
            return

        # If record doesn't exist, then failure
        if msg.content[1].name not in self.agent.storage:
            self.agent.send(failure_message_from_request(msg, "Not registered"))
            return

        # Otherwise perform the change
        self.agent.storage[msg.content[1].name] = msg.content[1]
        self.agent.send(inform_message_from_request(msg, msg.content))

    def _filter_descriptions(
        self, filter_description: DFAgentDescription
    ) -> List[DFAgentDescription]:
        """
        Filter all registered descriptions given the filter.

        :param filter_description: filter against registered descriptions will be tested
        """
        return list(
            filter(
                lambda e: _match_description(e, filter_description),
                self.agent.storage.values(),
            )
        )


def _match_description(
    description: DFAgentDescription, description_filter: DFAgentDescription
) -> bool:
    """
    Decide if the given `description` match the provided filter.

    :param description: description to be filtered
    :param description_filter: filter to test the description against
    :return: `True` if the description matches, `False` otherwise
    """
    return (
        (description_filter.name is None or description.name == description_filter.name)
        and (
            description_filter.services is None
            or set(description_filter.services) <= set(description.services)
        )
        and (
            description_filter.protocols is None
            or set(description_filter.protocols) <= set(description.protocols)
        )
        and (
            description_filter.ontologies is None
            or set(description_filter.ontologies) <= set(description.ontologies)
        )
        and (
            description_filter.languages is None
            or set(description_filter.languages) <= set(description.languages)
        )
        and (
            description_filter.lease_time is None
            or description_filter.lease_time == description.lease_time
        )
        and (
            description_filter.scope is None
            or description_filter.scope == description.scope
        )
    )


@dataclass(eq=True, frozen=True)
class ServiceDescription:
    """
    Description of each service registered with the DF.

    Available properties are:

    - **name**: name of the service
    - **type**: a string describing what kind of service it is. For example : "fipa-df"
    - **protocols**: protocols this agent supports. For example: "fipa-request"
    - **ontologies**: ontologies this agent understands. For example: "fipa-agent-management"
    - **languages**: languages this agent understands. For example: "fipa-sl0"
    - **ownership**: the owner of the service
    - **properties**: a set of properties

    It is part of the fipa-agent-management ontology. See http://fipa.org/specs/fipa00023/SC00023K.html.
    """

    name: str | None = None
    type: str | None = None
    protocols: Tuple[str, ...] = ()
    ontologies: Tuple[str, ...] = ()
    languages: Tuple[str, ...] = ()
    ownership: str | None = None
    properties: Tuple[Property, ...] = ()


@dataclass(eq=True, frozen=True)
class DFAgentDescription:
    """
    Description that can be registered with the DF service.

    - **name**: :class:`AID` of the agent
    - **services**: A set of services this agent supports
    - **protocols**: A list of interaction protocols this agent supports
    - **ontologies**:  A list of ontologies this agent supports. For example: "fipa-agent-management"
    - **languages**: A list of languages (for example "fipa-sl0") this agent knows.
    - **lease-time**: The duration or time at which the lease for this registration will expire
    - **scope**: Visibility of the record. Can be either local or global (default is global)

    This object is part of the fipa-agent-management ontology. See http://fipa.org/specs/fipa00023/SC00023K.html.
    """

    name: AID | None = None
    services: Tuple[ServiceDescription, ...] = ()
    protocols: Tuple[str, ...] = ()
    ontologies: Tuple[str, ...] = ()
    languages: Tuple[str, ...] = ()
    lease_time: datetime | None = None
    scope: str = "global"


@dataclass(eq=True, frozen=True)
class SearchConstraints:
    """
    This object is used to constraint the results of a search (either in the AMS or DF).

    It contains the following fields:

    - **max_depth**: if the DF agent supports federation, use this parameter to limit the request propagation.
    - **max_results**: limit the number of results returned by the request
    - **search_id**: assign a globally unique identifier for this search request

    Part of the fipa-agent-management ontology, see http://fipa.org/specs/fipa00023/SC00023K.html
    """

    max_depth: int | None = None
    max_results: int | None = None
    search_id: str | None = None



# coding: utf-8
"""
This example shows heavy exchanges between agents.

One sender with a bunch of receivers. Time elapsed between sending and receiving is displayed.
"""
from __future__ import annotations
import time

import piaf.comm
import piaf.agent
import piaf.behavior

from typing import Sequence


class HeavySenderBehavior(piaf.behavior.Behavior):
    """
    Behavior for the sender agent.

    It builds a message containing the current time and send it to other agents.
    """

    def __init__(self, agent: HeavySenderAgent):
        super().__init__(agent)
        self.last_call = None

    async def action(self):
        msg = (
            piaf.comm.ACLMessage.Builder()
            .performative(piaf.comm.Performative.INFORM)
            .receiver(self.agent.receivers)
            .content(time.time_ns())
            .build()
        )
        self.agent.send(msg)

    def done(self):
        """Infinite behavior."""
        return False


class ReceiveBehavior(piaf.behavior.Behavior):
    """
    Behavior for receivers.

    It unpacks the content and displays how much time elapsed between the mesage
    creation and its processing.
    """

    async def action(self):
        msg = await self.agent.receive()
        tm = (time.time_ns() - msg.acl_message.content) / 1_000_000
        self.logger.info(
            "[%s] Msg received in %ims.",
            self.agent.aid.short_name,
            tm,
        )

    def done(self):
        """Infinite behavior."""
        return False


class HeavySenderAgent(piaf.agent.Agent):
    """
    Agent in charge of sending messages.

    You must pass a list of aids identifying the receivers.
    """

    def __init__(self, aid, platform, receivers: Sequence[piaf.comm.AID]):
        """:param receivers: the receivers"""
        super().__init__(aid, platform)
        self.receivers = receivers

        self.add_behavior(HeavySenderBehavior(self))


class SimpleReceiverAgent(piaf.agent.Agent):
    """Receiver agent."""

    def __init__(self, aid, platform, *args, **kwargs):
        super().__init__(aid, platform, *args, **kwargs)

        self.add_behavior(ReceiveBehavior(self))


if __name__ == "__main__":
    import logging
    from typing import List

    import piaf.ptf
    import piaf.launcher
    import piaf.comm

    # Create the launcher
    launcher = piaf.launcher.PlatformLauncher("localhost")

    # Add receivers
    receivers: List[piaf.comm.AID] = []
    for i in range(100):
        receivers.append(piaf.comm.AID(f"rcv_{i}@localhost"))
        launcher.add_agent(
            piaf.launcher.AgentDescription(receivers[i].short_name, SimpleReceiverAgent)
        )

    # Add heavy sender
    launcher.add_agent(
        piaf.launcher.AgentDescription("sender", HeavySenderAgent, (receivers,))
    )

    # Configure logging to see things
    logging.getLogger().setLevel(logging.INFO)
    logging.getLogger().addHandler(logging.StreamHandler())

    # The program will run until you hit Ctrl+C
    launcher.run()



# coding: utf-8
"""
The module :mod:`piaf.behavior` contains everything related to behaviors.

Behaviors are small execution units that run "concurrently": several behaviors can be
active at the same time but only one is executed per agent.

Behaviors have access to the agent they are bounded to and are the preferred way to
manipulate it, like updating its knowledge, sending or receiving messages.
"""
from __future__ import annotations
import abc
import asyncio
import logging

from typing import Any, Callable, Dict, Type, Optional

import piaf.agent


__all__ = ["Behavior", "CyclicBehavior", "FSMBehavior"]


class Behavior(metaclass=abc.ABCMeta):
    """
    Base class user behaviors.

    When creating a behavior, you have to subclass :class:`Behavior` and provide at
    least  an implementation for the :meth:`action` method.

    You can also override the :meth:`done` method. This method is called when
    :meth:`action` method ends and according to the returned value, decide if
    the behavior ends or if the :meth:`action` method should be executed again.

    You should know that between two calls to :meth:`action`, a another
    behavior can be executed. It is also true when you are using the keyword
    ``await`` (mostlikely when you send or receive a message).

    You can access to the agent with the ``agent`` attribute.
    """

    def __init__(self, agent: piaf.agent.Agent):
        """
        Take an :class:`Agent` and construct a behavior with it.

        :param agent: the agent that will execute the behavior
        """
        self.agent = agent
        self.logger = logging.getLogger(type(self).__name__)

    async def run(self) -> None:
        """
        Run the behavior.

        This coroutine won't end until the :meth:`done` method returns
        ``False``. It also take care of the agent state: between two calls to
        method :meth:`action`, the agent's state is checked. The behavior will
        be paused if the agent's state is not ``AgentState.ACTIVE``.

        .. warning:: You should not override this method. To provide a custom body
                     to your behavior, implements the :meth:`action` method instead.
        """
        done = False
        while not done:
            async with self.agent.state_sync:
                await self.agent.state_sync.wait_for(
                    lambda: self.agent.state == piaf.agent.AgentState.ACTIVE
                )

            await self.action()
            done = self.done()
            if not done:
                await asyncio.sleep(0)  # Give a chance to another behavior

    @abc.abstractmethod
    async def action(self) -> None:
        """
        Body of any behavior.

        This method must be implemented. The :meth:`action` method is generally executed
        in a row, except if coroutines are awaited (mostlikely agent's
        :meth:`Agent.send`, :meth:`Agent.receive` or :meth:`Agent.receive_nowait`
        methods). Because of this, the :meth:`action` method shouldn't make expensive
        blocking operations, as it will slow down the entire platform.

        If you want to run expensive IO or CPU operations, use
        :func:`asyncio.run_in_executor` and await the result.
        """
        raise NotImplementedError()

    def done(self) -> bool:
        """
        Indicate wether or not this behavior is done.

        When this function `returns ``True``, the behavior's :meth:`run` method ends.
        Default implementation always return True.
        """
        return True

    def result(self) -> Any:
        """
        Get this behavior's result.

        Sometimes you might want to transmit a value. Rather than using an agent's
        attribute, you can use this method to fetch the result. Default is to return
        ``None``.
        """
        return None


class CyclicBehavior(Behavior, metaclass=abc.ABCMeta):
    """
    :class:`CyclicBehavior` are behaviors with a cyclic execution.

    Each time the :meth:`action` method execution finishes, a delay is applied before
    the next call. You can access (and modify) this delay using the ``delay`` attribute.

    Except this, it is a normal behavior.
    """

    def __init__(self, agent: piaf.agent.Agent, delay: float):
        """
        Create a :class:`CyclicBehavior`.

        The delay will pause the behavior between successive calls to method
        :meth:`action`.
        """
        super().__init__(agent)
        self.delay = delay

    def done(self) -> bool:
        """
        Indicate wether or not this behavior is done.

        When this function `returns ``True``, the behavior's :meth:`run` method ends.
        Default implementation always return False.
        """
        return False

    async def run(self) -> None:
        """
        Run the behavior.

        This coroutine won't end until the :meth:`done` method returns
        ``False``. It also take care of the agent state: between two calls to
        method :meth:`action`, the agent's state is checked. The behavior will
        be paused if the agent's state is ``AgentState.SUSPENDED``.

        The delay is applied between successive calls to :meth:`action`.

        .. warning:: You should not override this method. To provide a custom body
                     to your behavior, implements the :meth:`action` method instead.
        """
        done = False
        while not done:
            async with self.agent.state_sync:
                await self.agent.state_sync.wait_for(
                    lambda: self.agent.state == piaf.agent.AgentState.ACTIVE
                )
            await self.action()
            done = self.done()
            await asyncio.sleep(self.delay)


class _FSMState:
    """
    Used internally by the :class:`FSMBehavior` to represent states.

    This class is more like a data class that store information about states:

    ``name``: the state's name
    ``behavior``: the behavior class associated to the state
    ``transition``: a mapping function -> state
    ``final``: whether this state is a final state or not
    ``args``: a sequence of things that will be used to instantiate the behavior
    ``kwargs``: same as ``args`` but for keyword arguments

    """

    def __init__(
        self, name: str, behavior: Type[Behavior], args, kwargs, final: bool = False
    ):
        super().__init__()
        self.name = name
        self.behavior = behavior
        self.transitions: Dict[Callable[[Any], bool], _FSMState] = {}
        self.final = final
        self.args = args
        self.kwargs = kwargs

    def __eq__(self, value):
        return type(value) == type(self) and value.name == self.name

    def __hash__(self):
        return hash(self.name)


class FSMBehavior(Behavior):
    """
    Complex behavior that mimic a Final State Machine.

    To each state is associated a behavior that will be executed when the FSM enters in
    the associated state. Behaviors are instanciated each time.

    Like any FSM, you will need to identify which states are final and which is the
    initial state.

    Transitions made with functions with the following signature:
    ``Callable[[Any], bool]``. If the returned value is ``True``, then the FSM can
    transition to the targeted state. The function's argument will be fullfilled with
    the previous executed behavior result.

    Here is an example::

        # Create the behavior
        bhv = piaf.behavior.FSMBehavior(agent)

        # Four states: A, B, C and D
        # D is a final state
        bhv.add_state("A", ABehavior)
        bhv.add_state("B", BBehavior, args=(foo, ))
        bhv.add_state("C", CBehavior)
        bhv.add_state("D", DBehavior, args=(foo, bar), final=True)

        # Now A is our initial state
        bhv.set_initial_state("A")

        # Declare transitions.
        # Here our behaviors are returning a letter as result. We use it to make our
        # transition function.
        bhv.add_transition("A", "B", lambda r: r == "B")
        bhv.add_transition("A", "C", lambda r: r == "C")
        bhv.add_transition("B", "A", lambda r: r == "A")
        bhv.add_transition("C", "A", lambda r: r == "A")
        bhv.add_transition("C", "B", lambda r: r == "B")
        bhv.add_transition("B", "D", lambda r: r == "D")

    """

    def __init__(self, agent: piaf.agent.Agent):
        """Create a new :class:`FSMBehavior`. No states at the beginning."""
        super().__init__(agent)
        self._states: Dict[str, _FSMState] = {}
        self._initial: Optional[str] = None
        self._c_state: Optional[str] = None
        self._c_task: Optional[asyncio.Future[None]] = None
        self._last_result: Any = None

    def set_initial_state(self, name: str):
        """
        Set the provided state as the initial state.

        It will replace the previous selected one. Do not call this method once the
        behavior has been added to an agent.

        :param name: the state that will be the initial state
        """
        if name not in self._states:
            raise Exception()  # TODO: raise better exception

        self._initial = name

    def add_state(
        self,
        name: str,
        behavior: Type[Behavior],
        args=None,
        kwargs=None,
        final: bool = False,
    ):
        """
        Create a new state and associate the provided behavior.

        Names must be uniques for a given FSM. You must supply a behavior class (not
        an instance). When the state will be reached at execution time, the behavior
        will be instantiated and the FSM will wait until its completion to transition
        to the newt state.

        :param name: the state's name
        :param behavior: the behavior's class associated to this state
        :pram args: Optional. Should be a sequence of parameters required to your
                    behavior's instantiation (except for the ``agent`` parameter
                    that is automatically supplied)
        :param kwargs: same as ``args`` but for keywords arguments
        :param final: Is this state a final state ? Default is ``False``
        """
        if name in self._states:
            raise Exception()  # tODO: raise better exception

        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}

        self._states[name] = _FSMState(name, behavior, args, kwargs, final)

    def add_transition(self, from_: str, to_: str, func: Callable[[Any], bool]):
        """
        Add a transition between two states.

        The provided function will be used to decide if the transition is possible or
        not given the previous behavior result. Given a state, it shouldn't be possible
        to have two possible transition (meaning, two functions that returns true for
        two different states). This bad design can't be recognized, so extra
        carefullness is required here.

        :param from_: first state
        :param to_: second state
        :param func: transition function
        :raise ValueError: if either ``from_`` or ``to_`` state doesn't exist
        """
        try:
            from_s = self._states[from_]
        except KeyError:
            raise ValueError(f"Unknown state: {from_}")

        try:
            from_s.transitions[func] = self._states[to_]
        except KeyError:
            raise ValueError(f"Unknown state: {to_}")

    async def run(self) -> None:
        """
        Run the behavior.

        This coroutine won't end until the :meth:`done` method returns
        ``False``. It also take care of the agent state: between two calls to
        method :meth:`action`, the agent's state is checked. The behavior will
        be paused if the agent's state is ``AgentState.SUSPENDED``.

        A check will be performed just before running the behavior to ensure that:

        * There is an initial state
        * At least one final state is accessible
        * Not final states have at least one exit transition to another state
        * All  states are reachable

        The later will produce a warning each time a state can't be reached. The
        other three will raise an exception and prevent the behavior to be run.
        """
        self._check_fsm()
        await super().run()

    async def action(self) -> None:
        """
        :class:`FSMBehavior`'s body.

        Unlike :class:`Behavior`, there is no need to override this method. It will
        raise an exception if during execution there is no transition available and
        the current node is not final (FSM is stuck).
        """
        assert self._initial is not None

        # Compute current state
        if self._c_state is None:
            self._c_state = self._initial
        else:
            found = False
            for func, state in self._states[self._c_state].transitions.items():
                if func(self._last_result):
                    self._c_state = state.name
                    found = True
                    break
            if not found:
                raise Exception()  # TODO: better exception

        # Run behavior
        state = self._states[self._c_state]
        bhv = state.behavior(self.agent, *state.args, **state.kwargs)
        self._c_task = self.agent.add_behavior(bhv)

        # If task raises an exception, it will be propagated and stop the FSM
        await self._c_task
        self._last_result = bhv.result()

    def done(self) -> bool:
        """
        Indicate wether or not this behavior is done.

        When a final state is reached, this behavior will end (after the state's
        associated behavior execution).
        """
        assert self._c_state is not None
        assert self._c_task is not None

        return self._states[self._c_state].final

    def _check_fsm(self):
        if self._states.get(self._initial, None) is None:
            raise Exception()  # TODO: raise better exception

        visited = set()
        new_visited = {self._states[self._initial]}

        finals = set()
        if self._states[self._initial].final:
            finals.add(self._states[self._initial])

        while len(visited) != len(new_visited):
            visited = set(new_visited)
            for state in visited:
                for child in state.transitions.values():
                    new_visited.add(child)
                    if child.final:
                        finals.add(child)
                    elif not child.transitions:
                        raise Exception()  # TODO: raise better exception

        unreachable = set(self._states.values()).difference(visited)
        for state in unreachable:
            self.logger.warning(
                "[%s] Unreachable state: %s", self.agent.aid.short_name, state.name
            )

        if finals.issubset(unreachable):
            raise Exception()  # TODO: raise better exception


class SuicideBehavior(Behavior):
    """
    One-shot behavior killing the owner.

    A typicall usage is to associate this behavior to the final state of a :class:`FSMBehavior` instance. Although it
    only calls :meth:`Agent.quit()`, this behavior avoids duplication accross dependent projects.
    """

    async def action(self) -> None:
        """Make the owner quit as soon as possible."""
        await self.agent.quit()



# coding: utf-8
"""
The :mod:`piaf.agent` module contains everything you need to create your own agents.

According to FIPA specification:

.. epigraph::

    "An agent is a computational process that implements the autonomous, communicating
    functionality of an application. Agents communicate using an Agent Communication
    Language. An Agent is the fundamental actor on an [Agent Platform] which combines
    one or more service capabilities, as published in a service description, into a
    unified and integrated execution model. An agent must have at least one owner, for
    example, based on organisational affiliation or human user ownership, and an agent
    must support at least one notion of identity. This notion of identity is the Agent
    Identifier (AID) that labels an agent so that it  may be distinguished unambiguously
    within the Agent Universe. An agent may be registered at a number of transport
    addresses at which it can be contacted."

    -- `FIPA000023 <http://www.fipa.org/specs/fipa00023/SC00023K.html#_Toc75950978>`_

Currently, the Agent API supports:

* Communication with other Agents
* Agent IDentifier
* Multi transport addresses

It doesn't support:

* Agent ownership

"""
from __future__ import annotations

import asyncio
import copy
import datetime
import enum
import logging
from typing import TYPE_CHECKING, Iterable, MutableSequence, Sequence, Union

import piaf.comm
import piaf.comm.mts
import piaf.exceptions as ex
from piaf.events import AGENTS_PARENT_TOPIC, Event, Topic

if TYPE_CHECKING:
    import piaf.behavior
    import piaf.ptf

__all__ = ["Agent", "AgentState"]


class AgentState(enum.Enum):
    """
    Represents an agent state.

    According to FIPA specification (http://www.fipa.org/specs/fipa00023/), an agent
    can have 6 states:

    **INITIATED**
        Agent has been created or installed on the Agent Platform but is not invoked
        yet.
    **ACTIVE**
        Agent has been invoked and is now running.
    **SUSPENDED**
        Agent is suspended.
    **WAITING**
        Agent is waiting something.
    **TRANSIT**
        Agent is transiting between different platforms. Not supported for now.
    **UNKNOWN**
        Agent has been deleted or doesn't exist yet.

    More information in the specification.
    """

    def _generate_next_value_(name, start, count, last_values):  # noqa
        return name

    INITIATED = enum.auto()
    ACTIVE = enum.auto()
    SUSPENDED = enum.auto()
    WAITING = enum.auto()
    TRANSIT = enum.auto()
    UNKNOWN = enum.auto()


class Agent:
    """
    :class:`Agent` is the base class for all user agents.

    When creating an agent, you must provide this agent's identity and the platform
    facade that it will use to access vital functionalities like querying / updating
    the agent's state or communicating with other agents.

    You can use this class directly although your agent will be useless. We recommend
    subclassing.

    When subclassing the :class:`Agent` class, you can safely:

    * Extend the constructor to add new instructions
    * Extend all methods except the :meth: `run` method
    * Add custom methods

    Here is an example:

    .. code-block:: python

        from piaf.agent import Agent
        from piaf.launcher import AgentDescription, PlatformLauncher
        from piaf.comm.mts import MemoryMessageTransferHandler

        class MyAgent(Agent):
            def __init__(aid, platform, foo):
                '''Extra argument and attribute in the constructor.'''
                super().__init__(aid, platform)
                self.foo = foo

            def add_behavior(self, behavior):
                '''Extend method to add some logs.'''
                super().add_behavior(behavior)
                self.agent.logger.info(
                    "Added a behavior to agent %s", self.aid.short_name
                )

            def bar(self):
                '''Custom method.'''
                pass

        if __name__ == "__main__":
            # Create the launcher
            launcher = PlatformLauncher("localhost")

            # Add one agent to the simulation
            launcher.add_agent(AgentDescription("test", MyAgent, args=("bar",))

            # The program will run until you hit Ctrl+C
            launcher.run()

    """

    #: How much time (in seconds) the agent will sleep before checking if it is still
    #: in WAITING/SUSPENDED state.
    CLEANUP_DELAY = 1

    def __init__(
        self,
        aid: piaf.comm.AID,
        platform: piaf.ptf.AgentPlatformFacade,
    ):
        """
        Create a new Agent with the provided information.

        :param aid: this agent's identifier
        :param platform: the platform facade that proxies the platform where this agent is deployed
        """
        self.logger = logging.getLogger(type(self).__name__)
        self._aid = copy.deepcopy(aid)
        self._behaviors: MutableSequence[asyncio.Future[None]] = []
        self._platform = platform
        self._mailbox = piaf.comm.mts.MailBox()

    @property
    def state(self) -> AgentState:
        """
        Get this agent state.

        Readonly property that query the platform to get this agent's state.
        """
        return self._platform.get_state(self._aid)

    @property
    def state_sync(self):
        """Synchronization primitive."""
        return self._platform.get_state_condition(self._aid)

    @property
    def aid(self) -> piaf.comm.AID:
        """
        Get the agent's identifier.

        Please note that the retrieved object is a copy and can be safely modified
        without affecting the agent.
        Readonly property.
        """
        return copy.deepcopy(self._aid)

    @property
    def mailbox(self) -> piaf.comm.mts.MailBoxDelegate:
        """
        Get a reference to the internal's agent mailbox.

        You will get limited writing access.
        """
        return piaf.comm.mts.MailBoxDelegate(self._mailbox)

    async def wait(self) -> None:
        """Put this agent in WAITING state."""
        await self._platform.wait(self._aid)

    async def suspend(self) -> None:
        """Suspend this agent."""
        await self._platform.suspend(self._aid)

    async def quit(self) -> None:
        """Shutdown the agent as soon as possible."""
        await self._platform.quit(self._aid)

    def add_behavior(self, behavior: piaf.behavior.Behavior) -> asyncio.Future[None]:
        """
        Add a behavior to this agent.

        :param behavior: the behavior (an instance) you want to add.
        """
        task = asyncio.create_task(behavior.run())
        self._behaviors.append(task)
        return task

    async def run(self) -> None:
        """
        Coroutine in charge of running the agent.

        .. warning:: You should not override or extend this method or you agent might
                     not run correctly.
        """
        try:
            while self._running() and self._behaviors:
                # Wait until state is active or agent quits
                # If agent quits while being paused or suspended, we must unlock the agent
                async with self.state_sync:
                    await self.state_sync.wait_for(
                        lambda: self.state == AgentState.ACTIVE
                        or self.state == AgentState.UNKNOWN
                    )

                # If agent is active
                if self.state == AgentState.ACTIVE:
                    # Get done and pending tasks
                    done, pending = await asyncio.wait(
                        self._behaviors, timeout=self.CLEANUP_DELAY
                    )

                    # Display done behaviors with exception
                    self._display_bhv_failures(done)

                    # Update running behaviors list
                    self._behaviors = list(pending)

            # Inform the platform that we are quitting
            await self._platform.quit(self._aid)

        except asyncio.CancelledError:
            # Ignore error
            pass

        finally:
            # Cancel running behaviors
            if self._behaviors:
                for bhv in self._behaviors:
                    bhv.cancel()

                await asyncio.gather(*self._behaviors, return_exceptions=True)

            # Display behavior failures
            self._display_bhv_failures(self._behaviors)

            self.logger.info(
                "[%s]: If this is dying, I don't think much of it.",
                self._aid.short_name,
            )

    async def receive(
        self, template: piaf.comm.MessageTemplate = piaf.comm.MT_ALL()
    ) -> piaf.comm.Message:
        """
        Query this agent's mailbox to retrieve a message matching the provided template.

        Templates are applied to ACL Messages, not Envelopes.
        Note that this is a blocking receive, meaning that the coroutine ends when a
        message is found. The default template match any kind of messages.

        :param template: the message template used to search messages
        """
        return await self._mailbox.get(template)

    async def receive_nowait(
        self, template: piaf.comm.MessageTemplate = piaf.comm.MT_ALL()
    ) -> Union[None, piaf.comm.Message]:
        """
        Query this agent's mailbox to retrieve a message matching the provided template.

        Templates are applied to ACL Messages, not Envelopes.
        Contrary to :meth:`receive`, this method will return ``None`` if no matching
        message is found. The default template matches all messages.

        :param template: the message template used to search messages
        """
        return await self._mailbox.get_nowait(template)

    def send(self, message: piaf.comm.ACLMessage) -> None:
        """
        Send a message.

        The underlying implementation will use information provided in the message to
        contact receivers.

        :param message: the message to send
        """
        if message.receiver is None:
            raise ex.IllegalArgumentException("Message must have a receiver")

        if not isinstance(message.receiver, Sequence):
            to = (message.receiver,)
        else:
            to = message.receiver  # type: ignore

        message.sender = self.aid

        envelope = (
            piaf.comm.Envelope.Builder()
            .from_(self.aid)
            .to(to)
            .acl_representation("piaf.acl.repr.obj")
            .date(datetime.datetime.now())
            .build()
        )

        self._platform.send(piaf.comm.Message(envelope, message))

        # Fire event
        msg: piaf.comm.ACLMessage = copy.deepcopy(message)
        self._platform.evt_manager.publish_sync(
            Event(self.aid.name, "message_sending", msg),
            Topic.resolve_from_parent_topic(
                AGENTS_PARENT_TOPIC, f"{self.aid.short_name}.messages"
            ),
        )

    def _running(self) -> bool:
        return (
            self.state == AgentState.ACTIVE
            or self.state == AgentState.SUSPENDED
            or self.state == AgentState.WAITING
        )

    def _display_bhv_failures(self, tasks: Iterable[asyncio.Future[None]]):
        for task in filter(
            lambda e: not e.cancelled() and e.exception() is not None, tasks
        ):
            self.logger.exception(
                "%s: Behavior failure :",
                self.aid.short_name,
                exc_info=task.exception(),
            )



# coding: utf-8
"""
The :mod:`piaf.api.models` contains all `pydantic` models used to describe input and output data.

Every input models finish with the `In` suffix, while output models finish with the `Out` suffix.
"""
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field, conlist, validator

from piaf.agent import AgentState
from piaf.comm import Performative


class AgentPlatformModel(BaseModel):
    """
    Describe an incoming agent platform.

    It only contains one field named 'name', witch represents the platform's name to create.
    """

    name: str = Field(description="The name of the platform")

    class Config:
        """Extra configuration added to the model."""

        schema_extra = {"example": {"name": "my-awesome-platform"}}


class AgentCreationDescriptionModel(BaseModel):
    """
    Describe an incoming agent description.

    It should be supplied by the user in order to create an invoke an agent into a platform. The model contains four fields:

    - class_name: the fully qualified name of the agent's type
    - agent_name: the shortname of the agent to create
    - args: Optional, a sequence of arguments used to instantiate the agent
    - is_service: Default `True`, tells if the agent has a full access to the platform
    """

    class_name: str = Field(description="The fully qualified agent's class name")
    agent_name: str = Field(description="The desired shortname of the agent")
    args: Optional[List[Any]] = Field(
        default=None,
        description="A list of arguments passed to the agent's constructor",
    )
    is_service: bool = Field(
        default=True, description="Tell if the agent is a service or not"
    )

    class Config:
        """Extra configuration added to the model."""

        schema_extra = {
            "example": {
                "class_name": "custom.agents.MyCustomAgent",
                "agent_name": "Custom-1",
                "args": ["a-star", 3],
                "is_service": False,
            }
        }


class AIDModel(BaseModel):
    """
    Describe an agent identifier.

    The model contains four fields:

    - name: the full name of the agent, including the platform's name
    - addresses: a list of addresses used to reach the agent
    - resolvers: a list of naming resolvers
    """

    name: str = Field(
        description="The full name of the agent (format: shortname@ptf_name)",
        regex=r".*@.*",
    )
    addresses: conlist(str, unique_items=True) = Field(
        default=[],
        description="A set of addresses with which the agent can be reached.",
    )
    resolvers: conlist(AIDModel, unique_items=True) = Field(
        default=[],
        description="A set of agents that can resolve this agent's name.",
    )

    class Config:
        """Extra configuration added to the model."""

        schema_extra = {
            "example": {
                "name": "Custom-1@my-awesome-platform",
                "addresses": ["amqp://my-awesome-platform/acc"],
                "resolvers": [
                    {
                        "name": "ams@my-awesome-platform",
                        "addresses": ["amqp://my-awesome-platform/acc"],
                        "resolvers": [],
                    }
                ],
            }
        }


AIDModel.update_forward_refs()


class AMSAgentDescriptionModel(BaseModel):
    """
    Describe an agent when requested using the AMS.

    The model contains three fields:

    - aid: the agent's identifier
    - state: the state of the agent
    - owner: an optional owner of the agent
    """

    aid: AIDModel = Field(description="The AID of the described agent.")
    state: AgentState = Field(description="The state of the agent.")
    owner: Optional[str] = Field(default=None, description="The owner of the agent.")

    class Config:
        """Extra configuration added to the model."""

        schema_extra = {
            "example": {
                "aid": {
                    "name": "Custom-1@my-awesome-platform",
                    "addresses": [],
                    "resolvers": [],
                },
                "state": "active",
                "owner": None,
            }
        }


class ExceptionModel(BaseModel):
    """
    Describe an internal error to give clues about what went wrong.

    The model contains one field named `detail`, which describes the error.
    """

    detail: str = Field(
        description="An explanation text describing why the error happened."
    )

    class Config:
        """Extra configuration added to the model."""

        schema_extra = {"example": {"detail": "Unknown platform 'my-awesome-platform'"}}


class AgentStateModel(BaseModel):
    """
    Describe the state of an agent.

    It contains one field named `state` which must be either ACTIVE or SUSPENDED.
    """

    state: AgentState = Field("The state of the agent.")

    @validator("state")
    def restrict_state_values(cls, v):
        """
        Ensure the given value is either `AgentState.ACTIVE` or `AgentState.SUSPENDED`.

        :param cls: model class
        :param v: the value of the 'state' field
        :return: `v` if the value is valid
        :raise ValueError: v is not valid
        """
        if v not in (AgentState.ACTIVE, AgentState.SUSPENDED):
            raise ValueError("Must be either ACTIVE or SUSPENDED")
        return v

    class Config:
        """Extra configuration added to the model."""

        schema_extra = {"example": {"state": "ACTIVE"}}


class ACLMessageModel(BaseModel):
    """
    Describe an ACLMessage.

    It contains four fields:

    - `receivers`: a non-empty list of :class:`AIDModel` objects
    - `performative`: the message's performative
    - `conversation_id`: an optional conversation ID to track the conversation
    - `content`: a JSON-serializable message's content

    """

    receivers: conlist(AIDModel, min_items=1, unique_items=True) = Field(
        description="A set of AIDs, each one being the identity of a recipient."
    )
    performative: Union[Performative, str] = Field(
        description="The performative of the message."
    )
    conversation_id: Optional[str] = Field(
        default=None,
        description="A unique conversation ID shared by messages inside a conversation.",
    )
    content: Any = Field(
        description="The message's content, which must be a JSON-compatible one."
    )

    class Config:
        """Extra configuration added to the model."""

        schema_extra = {
            "example": {
                "receivers": [
                    {
                        "name": "Custom-2@my-awesome-platform",
                        "addresses": [],
                        "resolvers": [],
                    }
                ],
                "performative": "request",
                "conversation_id": "talk#3",
                "content": {"title": "Hello!", "repeat": 3},
            }
        }


class AgentMemoryModel(BaseModel):
    """
    The response model used when the client asks for a snapshot of an agent's memory.

    This model contains two fields:

    - `target`, which is the AID of the targeted agent
    - `memory`, which contains the memory snapshot
    """

    target: AIDModel
    memory: Dict[str, Any]

    class Config:
        """Extra configuration added to the model."""

        schema_extra = {
            "example": {
                "target": {
                    "name": "ams@localhost",
                    "shortname": "ams",
                    "addresses": [],
                    "resolvers": [],
                },
                "memory": {
                    "CLEANUP_DELAY": 1,
                    "CREATE_AGENT_FUNC": "create_agent",
                    "MODIFY_FUNC": "modify",
                    "SEARCH_FUNC": "search",
                    "aid": {
                        "name": "ams@localhost",
                        "shortname": "ams",
                        "addresses": [],
                        "resolvers": [],
                    },
                    "state": "ACTIVE",
                },
            }
        }



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['piaf',
 'piaf.api',
 'piaf.api.impl',
 'piaf.comm',
 'piaf.comm.mtp',
 'piaf.examples',
 'piaf.examples.two_platforms']

package_data = \
{'': ['*']}

install_requires = \
['python-dotenv>=0.21,<0.22']

extras_require = \
{'amqp-mtp': ['aiormq>=6,<7', 'yarl>=1,<2'],
 'webapi': ['aioredis>=2,<3', 'async-timeout>=4,<5', 'fastapi>=0.85,<0.86']}

setup_kwargs = {
    'name': 'piaf',
    'version': '0.2.0b3',
    'description': 'A FIPA-compliant Agent Platform written in python.',
    'long_description': '# Python Intelligent Agent Framework (piaf)\n\n![pipeline status](https://gitlab.com/ornythorinque/piaf/badges/master/pipeline.svg)\n![coverage report](https://gitlab.com/ornythorinque/piaf/badges/master/coverage.svg?job=test)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/piaf)\n![PyPI - License](https://img.shields.io/pypi/l/piaf)\n![PyPI](https://img.shields.io/pypi/v/piaf)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/piaf)\n\nThe aim of piaf is to provide a FIPA-compliant agent framework using Python. It uses **asyncio** to power agents.\n\n## Project status\n\n**The first official release is there!** After almost two years of development, I am now considering piaf is stable enough to start playing with it. I can\'t wait to see the amazing things you will do!\n\nSo, what is next ? Well there are still missing features in piaf and the next version will try to add some of them (I see you FIPA SL!).\n\n## Features\n\nAlthough piaf made some progress, it still needs some love to be fully compliant with the [FIPA specification](http://fipa.org/repository/standardspecs.html).\n\nWe provide some examples to help you understand what is possible to create with the current version, take a look at <https://gitlab.com/ornythorinque/piaf/-/tree/master/src/piaf/examples>.\n\n### Supported features\n\n- AMS (partial, only the query function)\n- DF\n- Communications within a platform\n- Communications between two **piaf platforms** (with some limitations)\n\n### Missing features\n\n- FIPA SL support (only plain Python objects are supported)\n- Federated DF\n- Name resolution\n- "Official" envelope representations (XML, bit-efficient) and MTPs (mainly HTTP, we don\'t plan to support IIOP)\n\n## Documentation\n\nThe full documentation (both user and API) is available here: <https://ornythorinque.gitlab.io/piaf>\nIt will teach you how to install and run your own agents.\n\n## Author(s)\n\n* ornythorinque (pierredubaillay@outlook.fr)\n',
    'author': 'Pierre DUBAILLAY',
    'author_email': 'pierredubaillay@outlook.fr',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gitlab.com/ornythorinque/piaf',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



# coding: utf-8
"""This module contains all the tools to work with events."""
from __future__ import annotations

import abc
import asyncio
from dataclasses import dataclass
from time import time
from typing import Any, Dict, Iterable, Set, Tuple


@dataclass(frozen=True, eq=True)
class Event:
    """
    The minimal structure of an event.

    It contains three fields:

    - `source`: from which component the event is emitted
    - `type`: what kind of event it is. Useful to know the content of the `data` field
    - `data`: the piece of data the event contains

    This class shouldn't be extended. Use the combination of the `type` and `data` field to extend the diversity of events.
    """

    source: str
    type: str
    data: Any


@dataclass(frozen=True, eq=True)
class EventRecord:
    """
    A tiny wrapper around an :class:`Event` that brings additional information.

    Each record contains three fields:

    - `event`: the :class:`Event` instance this record wraps
    - `timestamp`: when the event was emitted (second resolution)
    - `topics`: a tuple of topics where the event was broadcasted

    :class:`EventRecord` is what you manipulate in a custom :class:`Subscriber`.
    """

    event: Event
    timestamp: int
    topics: Tuple[Topic, ...]


class Topic:
    """
    A :class:`Topic` is the way to filter events.

    Rather than being a simple string, topics are hierarchical objects very much like loggers. Each topic, except the root one (`Topic(name="", parent=None)), have a parent :class:`Topic`.

    Combined with the :class:`Subscriber`, topics allow complex events broadcast scenarios.

    Topics are cached, meaning that if you create two topics, they will have at least the root topic in common (hierarchy is a tree).
    """

    DEFAULT_SEP = "."
    __CACHE: Set[Topic] = set()

    @classmethod
    def __from_cache(cls, name: str, parent: Topic | None) -> Topic | None:
        """
        Lookup the cache and try to get an existing :class:`Topic`.

        :param name: the topic's name
        :param parent: the parent of the topic
        :return: the :class:`Topic` instance or `None` if there is no such topic in the cache.
        """
        for topic in cls.__CACHE:
            if topic._name == name and topic._parent == parent:
                return topic
        return None

    @classmethod
    def from_str(cls, topic: str) -> Topic:
        """
        The right way to create topics.

        This method will create the full hierarchy of topics from the given string. Parent topics are separated by `.` (dots). For example, "parent.topic" will create three topics: `Topic(name="topic", parent=Topic(name="parent", parent=Topic(name="", parent=None)))`

        :param topic: the full name of the topic
        :return: the full topic's hierarchy
        """
        parts = topic.split(cls.DEFAULT_SEP)
        if parts[0] != "":
            parts.insert(0, "")  # Root topic

        prev = None
        for part in parts:
            prev = Topic(part, prev)
        return prev  # type: ignore

    @classmethod
    def resolve_from_parent_topic(cls, parent: Topic, topic: str) -> Topic:
        """
        Resolve the given str against the provided parent topic.

        This method will create the full hierarchy of topics from the given string, starting from parent. Parent topics are separated by `.` (dots). For example, "parent.topic" will create three topics: `Topic(name="topic", parent=Topic(name="parent", parent=Topic(name="", parent=None)))`

        :param parent: the parent topic
        :param topic: the full name of the child topic
        :return: the full topic's hierarchy
        """
        parts = topic.split(cls.DEFAULT_SEP)
        if parts[0] == "":
            return parent

        prev = parent
        for part in parts:
            prev = Topic(part, prev)
        return prev

    def __new__(cls, name: str, parent: Topic | None):
        """
        Override the default behavior of `__new__` by looking up the cache.

        If a matching topic already exist (same name and parent), then the cached object is returned. Otherwise the default behavior is called and a new instance is created.

        :param name: the topic's name
        :param parent: the topic's parent
        :return: a :class:`Topic` instance
        """
        topic = cls.__from_cache(name, parent)
        if topic is not None:
            return topic
        return super().__new__(cls)

    def __init__(self, name: str, parent: Topic | None) -> None:
        """
        Initialize the :class:`Topic` instance returned by :meth:`Topic.__new__`.

        If the provided instance is in the cache, initialization is skipped. :class:`Topic` creation should be done using the classmethod :meth:`Topic.from_str` rather than using the constructor.

        :param name: the topic's name
        :param parent: the topic's parent
        """
        try:
            # If the object is in the cache, skip initialization
            self in self.__CACHE
            return
        except AttributeError:
            # The object is not yet initialized and thus doesn't have the _name and the _parent attributes
            pass

        self._name = name
        self._parent = parent
        self.__CACHE.add(self)

    @property
    def name(self) -> str:
        """
        Get this topic's name.

        :return: the name
        """
        return self._name

    @property
    def parent(self) -> Topic | None:
        """
        Get this topic's parent.

        :return: the parent or `None` if this topic is the root topic
        """
        return self._parent

    def __str__(self) -> str:
        return f"{str(self.parent) if self.parent is not None else ''}{self.DEFAULT_SEP if self.parent is not None else ''}{self.name}"

    def __repr__(self) -> str:
        return f"Topic[parent={str(self.parent)}, name={self.name}"

    def __eq__(self, o: object) -> bool:
        return isinstance(o, Topic) and o.name == self.name and o.parent == self.parent

    def __hash__(self) -> int:
        return hash((self.name, self.parent))


PLATFORM_TOPIC = Topic.from_str("platform")
ACC_TOPIC = Topic.from_str("platform.acc")
AGENTS_PARENT_TOPIC = Topic.from_str("platform.agents")


class Subscriber(metaclass=abc.ABCMeta):
    """
    Abstract definition of a subscriber.

    Subscribers have only one method called :meth:`Subscriber.on_event`, which gets called when an event is published on a topic the subscriber is listening to.

    The subscription part is realized in the :class:`EventManager`.
    """

    @abc.abstractmethod
    async def on_event(self, event_record: EventRecord) -> None:
        """
        Called when an event is published on a topic this subscriber is listening to.

        :param event_record: the :class:`EventRecord` that wraps the published event.
        """
        raise NotImplementedError()

    async def close(self):
        """
        Close this subscriber by performing cleanup.

        Default implementation does nothing.
        """
        pass


class EventManager:
    """
    The heart of the EventAPI.

    The :class:`EventManager` holds the binding between topics and subscribers and also the publish/notify algorithm.
    """

    def __init__(self) -> None:
        """Initialize a new :class:`EventManager` instance with no topics nor subscribers."""
        self._subscribers: Dict[Topic, Set[Subscriber]] = {}
        self._tasks: Set[asyncio.Task] = set()

    async def publish(self, event: Event, topics: Topic | Iterable[Topic]) -> None:
        """
        Publish the given event on the given topics.

        The notification algorithm is the following:

        1. Topics are expanded to get the full hierarchy of topics
        2. For each topic in the expanded set, get the registered subscribers
        3. For each (unique) subscriber, call the :meth:`Subscriber.on_event` method

        :param event: the event that is being published
        :param topics: either a :class:`Topic` or a list of topics where the event is published
        """
        if isinstance(topics, Topic):
            topics = [topics]

        # Expand topics to get the full hierarchy
        expanded_topics: Set[Topic] = set()
        for topic in topics:
            expanded_topics.add(topic)
            parent = topic.parent
            while parent is not None:
                expanded_topics.add(parent)
                parent = parent.parent

        # Retrieve subscribers
        subscribers: Set[Subscriber] = set()
        for topic in expanded_topics:
            subscribers |= self._subscribers.get(topic, set())

        # Notify subscribers
        record = EventRecord(event, int(time()), tuple(topics))
        for subscriber in subscribers:
            await subscriber.on_event(record)

    def publish_sync(self, event: Event, topics: Topic | Iterable[Topic]) -> None:
        """
        Synchronous version of :meth:`EventManager.publish`.

        :param event: the event that is being published
        :param topics: either a :class:`Topic` or a list of topics where the event is published
        """
        task = asyncio.create_task(self.publish(event, topics))
        self._tasks.add(task)

        async def _cleanup():
            await task
            self._tasks.discard(task)

        asyncio.create_task(_cleanup())

    def subscribe_to(
        self, subscriber: Subscriber, topics: Topic | Iterable[Topic]
    ) -> None:
        """
        Bind the given subscriber to the provided list of topics.

        After this call, the subscriber will get notified when an event is published on one of the given topics.

        :param subscriber: the subscriber to bind
        :param topics: either a :class:`Topic` or a list of topics that will be bound to the subscriber
        """
        if isinstance(topics, Topic):
            topics = [topics]

        for topic in topics:
            subscribers = self._subscribers.setdefault(topic, set())
            subscribers.add(subscriber)

    async def close(self) -> None:
        """
        Close the :class:`EventManager` and give a chance to subscribers to cleanup.
        """
        # Wait until synchronous publishes are done
        if self._tasks:
            await asyncio.wait(self._tasks, return_when=asyncio.ALL_COMPLETED)

        # Close all subscribers
        all_subscribers = set()
        for subscribers in self._subscribers.values():
            all_subscribers.update(subscribers)
        map(Subscriber.close, all_subscribers)
